import unittest

from artap.problem import Problem

from artap.algorithm_bayes import Bayes, BayesianOptimization

import numpy as np


# PBOUNDS = {'p1': (-10, 10), 'p2': (-10, 10)}
PBOUNDS = {'x': (-2, 10)}


def target_func(x):
    # val = sum(x)
    # val = -Booth.eval(x)
    val = np.exp(-(x - 2)**2) + np.exp(-(x - 6)**2/10) + 1/ (x**2 + 1)
    # print(val, x)
    return val


# class MyProblemOne(Problem):
#     """ Describe simple one objective optimization problem. """
#
#     def set(self):
#         self.name = "TestBayesOptParallel"
#         self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-2, 10]}]
#         self.costs = [{'name': 'F'}]
#
#     def evaluate(self, individual):
#         x = individual.vector
#         return -(np.exp(-(x - 2)**2) + np.exp(-(x - 6)**2/10) + 1 / (x**2 + 1))


# class MyProblem(Problem):
#     """ Describe simple one objective optimization problem. """
#     def set(self):
#         self.name = "TestBayesOptSerial"
#         self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
#                       {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
#         self.costs = [{'name': 'F'}]
#
#     def evaluate(self, individual):
#         return [Booth.eval(individual.vector)]


# class TestBayesOptimization(unittest.TestCase):
#     """ Tests simple one objective optimization problem."""
#
#     def test_bayesoptimization(self):
#         optimizer = BayesianOptimization(f=target_func,
#                                          pbounds=PBOUNDS,
#                                          random_state=np.random.RandomState(1))
#
#         optimizer.maximize(init_points=5, n_iter=30, acq="ucb", kappa=10.0)
#         print("maximum")
#         print(optimizer.max)


if __name__ == '__main__':
    unittest.main()



from .individual import Individual


class Population:
    def __init__(self, individuals: list = None, archives: list = None):
        if individuals is None:
            self.individuals = []
        else:
            self.individuals = individuals.copy()

        if archives is None:
            self.archives = []
        else:
            self.archives = archives.copy()

        self.pareto_vectors: list = []
        self.pareto_costs: list = []

    def __str__(self):
        string = "individuals: {}\n".format(len(self.individuals))
        for individual in self.individuals:
            string += "\t{},\n".format(individual.__str__())
        return string

    def __eq__(self, other):
        if not isinstance(other, Population):
            return NotImplemented

        return self.individuals == other.individuals and self.archives == other.archives and self.pareto_vectors == other.pareto_vectors and self.pareto_costs == other.pareto_costs


from multiprocessing import Process, Event
import os
import time
from enum import Enum
from sqlitedict import SqliteDict


class Timer(Process):
    def __init__(self, interval, function, args=[], kwargs={}):
        super(Timer, self).__init__(*args, **kwargs)
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.finished = Event()

        self.daemon = True

    def cancel(self):
        self.finished.set()

    def run(self):
        while not self.finished.is_set():
            self.function(*self.args, **self.kwargs)
            self.finished.wait(self.interval)
        self.finished.set()


class FileMode(Enum):
    READ = 0
    WRITE = 1
    REWRITE = 2


class FileDataStore:
    def __init__(self, problem, database_name, mode=FileMode.WRITE):
        self._should_continue = False
        self.is_running = False

        self.problem = problem
        self.database_name = database_name
        self.mode = mode

        if not self.database_name:
            raise RuntimeError("FileDataStoreCacheThread: database name is empty.")

        elif self.mode == FileMode.REWRITE:
            if os.path.exists(database_name):
                os.remove(database_name)

        self.db = SqliteDict(self.database_name, autocommit=True)
        if self.mode == FileMode.WRITE:
            if os.path.exists(self.database_name):
                statinfo = os.stat(self.database_name)
                if statinfo.st_size == 0:
                    # raise RuntimeError("FileDataStoreCacheThread: database file already exists. Mode is WRITE (for automatic rewrite you can use REWRITE mode.")
                    self.read_from_datastore()
                else:
                    self._create_structure()
            else:
                self._create_structure()
        elif self.mode == FileMode.REWRITE:
            self._create_structure()
        elif self.mode == FileMode.READ:
            self.read_from_datastore()

        self.delay = 2.0
        self.start()

    def _handle_target(self):
        self.is_running = True
        self.sync()
        self.is_running = False

    def _start_timer(self):
        if self._should_continue:
            # if self.timer is not None:
            #     self.timer.cancel()
            #     del self.timer
            #     self.timer = None
            timer = Timer(self.delay, self._handle_target)
            timer.start()

    def _create_structure(self):
        self.db["name"] = self.problem.name
        self.db["description"] = self.problem.description
        self.db["parameters"] = self.problem.parameters
        self.db["costs"] = self.problem.costs

        self.db["populations"] = []

    def read_from_datastore(self):
        self.problem.name = self.db["name"]
        self.problem.description = self.db["description"]
        self.problem.parameters = self.db["parameters"]
        self.problem.costs = self.db["costs"]

        self.problem.populations = self.db["populations"]

    def start(self):
        if not self._should_continue and not self.is_running:
            self._should_continue = True
            self._start_timer()

    def destroy(self):
        self._should_continue = False

        self.sync()
        self.db.close()

    def sync(self):
        if self.mode == FileMode.WRITE or self.mode == FileMode.REWRITE:
            # print("sync: {} - {}".format(self.problem, len(self.problem.populations)))
            if len(self.problem.populations) > 0 and self.db.conn is not None:
                # print("SYNC - 3")
                self.problem.logger.info("Caching to disk {} populations.".format(len(self.problem.populations)))

                self.db["populations"] = self.problem.populations



"""
Implementation of the Semi analytical TEAM Benchmark problem in Artap.

The ProblemAnalytical class contains the description oe optimization problem, 10 design parameters, 1-3 goal functions,
the initialization of the database and the functions for the realized semi-analytical calculation.

The detailed description of the analytical model can be founc here: doi....
"""

import math
import pylab as pl

from scipy import integrate
from artap.problem import Problem
from artap.individual import Individual
from artap.algorithm_nlopt import NLopt, LN_BOBYQA
from artap.datastore import FileDataStore


class ProblemAnalytical(Problem):
    def set(self):

        self.name = "Semi-Analytical solution"

        # The problem has 10 parameters, because the optimized coil has 20 turns, but during the solution, only the
        # symmetrycal cases has been assumed. Other simplification is that only the x coordinate of the turn can be
        # changed.
        #
        # Note: 'initial_value' should be given for the nl_opt and the bayesopt methods, not sufficient for genetic
        #        algorithms
        self.parameters = [{'name': 'x1', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x2', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x3', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x4', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x5', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x6', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x7', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x8', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x9', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x10', 'initial_value': 0.01, 'bounds': [5.01e-3, 50e-3]}]

        # The three, separate optimization functions and the direction of the optimization
        # is set to minimization. It is also possible to use the maximize keyword.
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

        # Initialize a database to store the data into the given space
        self.data_store = FileDataStore(self, database_name="/tmp/team_multi.db", mode="write")

    def intl22(self, R2, R, dZ, phi):
        return math.sqrt(R2 ** 2 + R ** 2 - 2.0 * R2 * R * math.cos(phi) + dZ ** 2)

    def intg(self, R2, R, dZ):
        # div J = 0 - nonconstant current density
        f = lambda phi: math.log(R2 - R * math.cos(phi) + self.intl22(R2, R, dZ, phi)) * math.cos(phi)
        # constant current density
        # f = lambda phi: R * math.cos(phi)**2 * math.log(R2 - R * math.cos(phi) + self.intl22(R2, R, dZ, phi)) + math.cos(phi) * self.intl22(R2, R, dZ, phi)
        return integrate.quad(f, 0, 2.0 * math.pi, epsabs=1e-3, epsrel=1e-3)[0]

    def inth(self, R2, R, dZ):
        # div J = 0 - nonconstant current density
        f = lambda phi: - math.log(dZ + self.intl22(R2, R, dZ, phi))
        return integrate.quad(f, 0, 2.0 * math.pi, epsabs=1e-3, epsrel=1e-3)[0]

    def integral(self, rc, zc, R, Z):

        w = 0.001
        h = 0.0015

        R1 = rc
        R2 = rc + w
        Z1 = zc
        Z2 = zc + h

        mu0 = 4.0 * math.pi * 1e-7
        Jext = 2e6
        # div J = 0 - nonconstant current density
        C = mu0 * Jext * w * h / (4 * math.pi * (Z2 - Z1) * math.log(R2 / R1))
        # constant current density
        # C = mu0 * Jext / (4.0 * math.pi)

        # upper coil
        Bru = C * (self.intg(R2, R, Z2 - Z) - self.intg(R2, R, Z1 - Z) - self.intg(R1, R, Z2 - Z) + self.intg(R1, R,
                                                                                                              Z1 - Z))
        Bzu = C * (self.inth(R2, R, Z2 - Z) - self.inth(R2, R, Z1 - Z) - self.inth(R1, R, Z2 - Z) + self.inth(R1, R,
                                                                                                              Z1 - Z))

        # lower coil
        Brl = C * (self.intg(R2, R, -Z1 - Z) - self.intg(R2, R, -Z2 - Z) - self.intg(R1, R, -Z1 - Z) + self.intg(R1, R,
                                                                                                                 -Z2 - Z))
        Bzl = C * (self.inth(R2, R, -Z1 - Z) - self.inth(R2, R, -Z2 - Z) - self.inth(R1, R, -Z1 - Z) + self.inth(R1, R,
                                                                                                                 -Z2 - Z))

        return [Bru + Brl, Bzu + Bzl]

    def integral_all(self, R, Z, x):
        Br = 0.0
        Bz = 0.0

        for k in range(0, 9):
            rc = x[k]
            zc = k * 1.5e-3

            B = self.integral(rc, zc, R, Z)

            Br = Br + B[0]
            Bz = Bz + B[1]

        return [Br, Bz]

    def evaluate(self, individual):
        B0 = 2e-3

        dxy = 0.5e-3
        nx = 8
        ny = 8
        dx = (5e-3 - dxy) / (nx - 1)
        dy = (5e-3 - dxy) / (ny - 1)
        f1 = 0.0
        f2 = 0.0
        f3 = 0.0

        for i in range(0, nx):
            xx = dxy + i * dx
            for j in range(0, ny):
                yy = dxy + j * dy

                [Br, Bz] = self.integral_all(xx, yy, individual.vector)
                Bp1s = math.sqrt((Br - 0.0) ** 2 + (Bz - B0) ** 2)
                f1 = max(f1, Bp1s)

                dxsi = 0.5e-3
                [Brp, Bzp] = self.integral_all(xx + dxsi, yy, individual.vector)
                [Brm, Bzm] = self.integral_all(xx - dxsi, yy, individual.vector)
                Bp2 = math.sqrt((Brp - Br) ** 2 + (Bzp - Bz) ** 2) + math.sqrt((Brm - Br) ** 2 + (Bzm - Bz) ** 2)
                f3 = max(f2, Bp2)

        f2 = sum(individual.vector) * 1e3

        return [f1,f2]


def optim_single():
    problem = ProblemAnalytical()

    # optimization
    algorithm = NLopt(problem)
    algorithm.options['algorithm'] = LN_BOBYQA
    algorithm.options['n_iterations'] = 100

    algorithm.run()


def check_analytical():
    """
    Checks the functionality of the semi analytical function.
    :return:
    """
    vec = [0.00808, 0.0149, 0.00674, 0.0167, 0.00545, 0.0106, 0.0117, 0.0111, 0.01369, 0.00619]
    # vec = [5e-3] * 10
    # vec = [10e-3] * 10
    ind = Individual(vec)
    problem_analytical = ProblemAnalytical()
    a = problem_analytical.evaluate(ind)

    print("anal ", a)


def check_plot():
    """
    This function makes a plot from the solution along the z axis at r = 0.00619 m.
    Compares the function of the semi-analytical integral values with 20 different points, which calculated by Agros.
    """

    problem_analytical = ProblemAnalytical()

    # Agros
    za = pl.linspace(0, 0.03, 20)
    Bra = [1.50901e-10, -5.11533e-6, -1.11251e-5, -1.91503e-5, -3.08628e-5, -4.90497e-5, -7.81889e-5, -0.000124345,
           -0.000167617, -1.92214e-6, 0.000169002, 0.000127021, 8.07604e-5, 5.16543e-5, 3.38259e-5, 2.27073e-5,
           1.56076e-5, 1.09697e-5, 7.88858e-6, 5.78184e-6]
    Bza = [3.22309e-5, 3.34485e-5, 3.72951e-5, 4.43861e-5, 5.60296e-5, 7.48665e-5, 0.000107128, 0.000170325,
           0.000319292, 0.000505151, 0.000322344, 0.000170171, 0.000105039, 7.12451e-5, 5.08364e-5, 3.73934e-5,
           2.81052e-5, 2.14857e-5, 1.66542e-5, 1.30694e-5]

    z = pl.linspace(0, 0.03, 100)
    Br = []
    Bz = []

    for i in range(len(z)):
        r1 = 0.00619
        z1 = 9 * 1.5e-3
        B = problem_analytical.integral(r1, z1, 0.005, z[i])

        Br.append(B[0])
        Bz.append(B[1])

    # Plot the comparison of the analytical and the agros solution
    pl.subplot(2, 1, 1)
    pl.plot(z, Bz, label="Semi-analytical method")
    pl.plot(za, Bza, 'x', label="Agros")
    pl.ylabel("Bz [T]")
    pl.legend()
    pl.subplot(2, 1, 2)
    pl.plot(z, Br, label="Semi-analytical method")
    pl.plot(za, Bra, 'x', label="Agros")
    pl.ylabel("Br [T]")
    pl.xlabel('r [m]')
    pl.show()


if __name__ == "main":
    # check_analytical()
    check_plot()
    optim_single()



import unittest
import time
import random
import threading
import rpyc

from artap.algorithm_sweep import SweepAlgorithm
from artap.problem import Problem
from artap.operators import LHSGenerator
from artap.monitor import MONITOR_PORT


class SleepProblem(Problem):
    def set(self, **kwargs):
        self.name = 'Sleep Problem'
        self.parameters = [{'name': 'x', 'bounds': [-5., 5.]}]

        # single objective problem
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'}]

    def evaluate(self, x):
        time.sleep(random.randrange(1, 3))

        return [x.vector[0]**2]


class TestServer(unittest.TestCase):
    def run_test(self):
        generator = LHSGenerator(self.problem.parameters)
        generator.init(5)

        algorithm = SweepAlgorithm(self.problem, generator=generator)
        algorithm.options['verbose_level'] = 0
        algorithm.options['max_processes'] = 3
        algorithm.options['n_iterations'] = 5
        algorithm.run()

    def test_run(self):
        self.problem = SleepProblem()

        # execute algorithm in different thread
        thread = threading.Thread(target=self.run_test, args=())
        thread.start()

        # connect to server
        conn = rpyc.classic.connect("localhost", self.problem.monitor_service.port)
        while thread.is_alive():
            # print(len(conn.root.populations()))
            populations = conn.root.populations()
            if len(populations) == 1:
                for individual in populations[0].individuals:
                    # print("individual = {}".format(individual))
                    pass

            time.sleep(0.1)

        populations = conn.root.populations()
        self.assertEqual(len(populations[0].individuals), 4)


class TestServer(unittest.TestCase):
    def run_test(self, n):
        problem = SleepProblem()
        problem.name = "SleepProblem{}".format(n)
        generator = LHSGenerator(problem.parameters)
        generator.init(1)

        algorithm = SweepAlgorithm(problem, generator=generator)
        algorithm.options['verbose_level'] = 0
        algorithm.options['max_processes'] = 1
        algorithm.options['n_iterations'] = 1
        algorithm.run()

    def test_run(self):
        ok = False

        n = 4
        for i in range(n):
            # execute algorithm in different thread
            thread = threading.Thread(target=self.run_test, args=(i, ))
            thread.start()
            time.sleep(0.2)

        port = MONITOR_PORT
        # connect to server
        i = 0
        while i < n:
            try:
                conn = rpyc.classic.connect("localhost", port)
                # print("Connected to {} at port {}".format(conn.root.problem.name, port))
                i = i + 1
            except Exception as e:
                # print(e)
                pass

            port = port + 1

            # emergency exit
            if port > MONITOR_PORT + n + 100:
                break

        self.assertTrue(i == n)


if __name__ == '__main__':
    unittest.main()



"""
 Module is dedicated to describe optimization problem.
"""

from artap.datastore import FileDataStore, FileMode
from artap.utils import ConfigDictionary
from artap.surrogate import SurrogateModelEval
from artap.monitor import MonitorService
from abc import abstractmethod

import logging
import datetime
import tempfile
import os
import sys
import shutil
import atexit

CRITICAL = logging.CRITICAL
FATAL = logging.FATAL
ERROR = logging.ERROR
WARNING = logging.WARNING
WARN = logging.WARN
INFO = logging.INFO
DEBUG = logging.DEBUG
NOTSET = logging.NOTSET

_log_level = [CRITICAL, ERROR, WARNING, INFO, DEBUG]


class Problem:
    """ The Class Problem Is a main class which collects information about optimization task """

    # __is_frozen = False

    def __init__(self, **kwargs):

        self.options = ConfigDictionary()
        # options
        self.options.declare(name='calculate_gradients', default=False,
                             desc='calculate gradient for individuals')
        self.options.declare(name='log_level', default=logging.INFO, values=_log_level,
                             desc='Log level')
        self.options.declare(name='log_file_handler', default=False,
                             desc='Enable file handler')
        self.options.declare(name='log_db_handler', default=True,
                             desc='Enable DB handler')
        self.options.declare(name='log_console_handler', default=True,
                             desc='Enable console handler')
        self.options.declare(name='time_out', default=600,
                             desc='Maximal time for calculation')
        self.options.declare(name='save_data_files', default=False,
                             desc='Saving data from computation')

        # tmp name
        d = datetime.datetime.now()
        ts = d.strftime("{}-%f".format(self.__class__.__name__))

        # create logger
        self.logger = logging.getLogger(ts)
        self.logger.setLevel(self.options['log_level'])

        # create formatter
        self.formatter = logging.Formatter('%(asctime)s (%(levelname)s): '
                                           '%(name)s - %(funcName)s (%(lineno)d) - %(message)s')

        if self.options['log_console_handler']:
            # create console handler and set level to debug
            stream_err_handler = logging.StreamHandler()
            stream_err_handler.setLevel(logging.DEBUG)
            # add formatter to StreamHandler
            stream_err_handler.setFormatter(self.formatter)
            # add StreamHandler to logger
            self.logger.addHandler(stream_err_handler)

            # stream_std_handler = logging.StreamHandler()
            # stream_std_handler.setLevel(logging.INFO)
            # # add formatter to StreamHandler
            # stream_std_handler.setFormatter(self.formatter)
            # # add StreamHandler to logger
            # self.logger.addHandler(stream_std_handler)

        for h in list(self.logger.handlers):
            print(h)

        self.logger.debug("START")

        self.name: str = str()
        self.description: str = str()
        self.parameters: dict = dict()
        self.costs: dict = dict()

        # populations
        self.populations = []

        self.data_store = None
        self.monitor_service = MonitorService(self)
        self.executor = None

        self.output_files = None

        self.working_dir = tempfile.gettempdir() + os.sep + "artap-{}".format(ts) + os.sep
        os.mkdir(self.working_dir)

        # surrogate model (default - only simple eval)
        self.surrogate = SurrogateModelEval(self)
        self.signs = []

        # self._freeze()
        self.set(**kwargs)
        for cost in self.costs:
            if 'criteria' in cost:
                if cost['criteria'] == 'minimize':
                    self.signs.append(1)
                else:
                    self.signs.append(-1)
            else:
                self.signs.append(1)

        # clean up
        atexit.register(self.cleanup)

    def cleanup(self):
        if self.data_store:
            self.data_store.destroy()

        if self.monitor_service.server is not None:
            self.monitor_service.server.close()
            del self.monitor_service.server
            self.monitor_service.server = None

        if self.working_dir.startswith(tempfile.gettempdir() + os.sep + "artap-"):
            shutil.rmtree(self.working_dir)

    @abstractmethod
    def set(self, **kwargs):
        pass

    def parameters_len(self):
        return len(self.parameters)

    # search for and defines the initial values for the optimization problem
    def get_initial_values(self):
        values = []
        for parameter in self.parameters:
            if 'initial_value' in parameter:
                values.append(parameter['initial_value'])
            else:
                values.append(0)
        return values

    def get_parameter_types(self):
        # TODO: is it important? or a typecheck is enough?
        p_types = []
        for parameter in self.parameters:

            if 'parameter_type' in parameter:
                if parameter['parameter_type'].lower == 'real':
                    p_types.append('real')

                if parameter['parameter_type'].lower == 'integer':
                    p_types.append('integer')

                if parameter['parameter_type'].lower == 'boolean':
                    p_types.append('boolean')

            else:
                p_types.append('real')

        return p_types

    @abstractmethod
    def evaluate(self, x):
        """ :param x: individual """
        pass

    def evaluate_constraints(self, x):
        """ :param x: individual """
        pass

    def __setattr__(self, key, value):
        # if self.__is_frozen and not hasattr(self, key):
        #     raise TypeError(" %r is a frozen class" % self)
        object.__setattr__(self, key, value)

        # working dir must be set
        if hasattr(self.options, 'log_file_handler'):
            if self.working_dir:
                # create file handler and set level to debug
                file_handler = logging.FileHandler(self.working_dir + "/data.log")
                file_handler.setLevel(logging.DEBUG)
                # add formatter to FileHandler
                file_handler.setFormatter(self.formatter)
                # add FileHandler to logger
                self.logger.addHandler(file_handler)

    # def _freeze(self):
    #     self.__is_frozen = True


class ProblemViewDataStore(Problem):
    def __init__(self, database_name):
        super().__init__()

        self.data_store = FileDataStore(self, database_name=database_name, mode=FileMode.READ)

    def set(self, **kwargs):
        pass

    def evaluate(self, x):
        pass



import collections
from numpy.random import normal
from random import random


def flatten(l):
    output = []
    if isinstance(l, collections.Iterable):
        for item in l:
            if isinstance(item, collections.Iterable):
                output.append(flatten(item))
            else:
                output.append(item)
    else:
        output.append(l)

    return output


class ConfigDictionary(object):
    """
    Attributes
    ----------
    _dict : dict of dict
        Dictionary of entries. Each entry is a dictionary consisting of value, values, types, desc, lower, and upper.
    _read_only : bool
        If True, setting (via __setitem__ or update) is not permitted.
    """

    def __init__(self, read_only=False):
        """
        Initialize all attributes.

        Parameters
        ----------
        read_only : bool
            If True, setting (via __setitem__ or update) is not permitted.
        """
        self._dict = {}
        self._read_only = read_only

    def _assert_valid(self, name, value):
        """
        Check whether the given value is valid, where the key has already been declared.

        The optional checks consist of ensuring: the value is one of a list of acceptable values,
        the type of value is one of a list of acceptable types, value is not less than lower,
        value is not greater than upper, and value satisfies is_valid.

        Parameters
        ----------
        name : str
            The key for the declared entry.
        value : object
            The default or user-set value to check for value, type, lower, and upper.
        """
        meta = self._dict[name]
        values = meta['values']
        types = meta['types']
        lower = meta['lower']
        upper = meta['upper']
        is_valid = meta['is_valid']

        if not (value is None and meta['allow_none']):
            # If only values is declared
            if values is not None:
                if value not in values:
                    raise ValueError("Entry '{}'\'s value is not one of {}".format(name, values))
            # If only types is declared
            elif types is not None:
                if not isinstance(value, types):
                    raise TypeError("Entry '{}' has the wrong type ({})".format(name, types))

            if upper is not None:
                if value > upper:
                    msg = "Value of {} exceeds maximum of {} for entry 'x'"
                    raise ValueError(msg.format(value, upper))
            if lower is not None:
                if value < lower:
                    msg = "Value of {} exceeds minimum of {} for entry 'x'"
                    raise ValueError(msg.format(value, lower))

        # General function test
        if is_valid is not None and not is_valid(value):
            raise ValueError("Function is_valid returns False for {}.".format(name))

    def declare(self, name, default, values=None, types=None, desc='',
                upper=None, lower=None, is_valid=None, allow_none=False):
        r"""
        Declare an option.

        The value of the option must satisfy the following:
        1. If values only was given when declaring, value must be in values.
        2. If types only was given when declaring, value must satisfy isinstance(value, types).
        3. It is an error if both values and types are given.

        Parameters
        ----------
        name : str
            Name of the option.
        default : object or Null
            Optional default value that must be valid under the above 3 conditions.
        values : set or list or tuple or None
            Optional list of acceptable option values.
        types : type or tuple of types or None
            Optional type or list of acceptable option types.
        desc : str
            Optional description of the option.
        upper : float or None
            Maximum allowable value.
        lower : float or None
            Minimum allowable value.
        is_valid : function or None
            General check function that returns True if valid.
        allow_none : bool
            If True, allow None as a value regardless of values or types.
        """

        if values is not None and not isinstance(values, (set, list, tuple)):
            raise TypeError("'values' must be of type None, list, or tuple - not %s." % values)
        if types is not None and not isinstance(types, (type, set, list, tuple)):
            raise TypeError("'types' must be None, a type or a tuple  - not %s." % types)

        if types is not None and values is not None:
            raise RuntimeError("'types' and 'values' were both specified for option '%s'." %
                               name)

        self._dict[name] = {
            'value': default,
            'values': values,
            'types': types,
            'desc': desc,
            'upper': upper,
            'lower': lower,
            'is_valid': is_valid,
            'allow_none': allow_none
        }

        # check for validity
        self._assert_valid(name, default)

    def update(self, in_dict):
        """
        Update the internal dictionary with the given one.

        Parameters
        ----------
        in_dict : dict
            The incoming dictionary to add to the internal one.
        """
        for name in in_dict:
            self[name] = in_dict[name]

    def __iter__(self):
        """
        Provide an iterator.

        Returns
        -------
        iterable
            iterator over the keys in the dictionary.
        """
        return iter(self._dict)

    def __contains__(self, key):
        """
        Check if the key is in the local dictionary.

        Parameters
        ----------
        key : str
            name of the entry.

        Returns
        -------
        boolean
            whether key is in the local dict.
        """
        return key in self._dict

    def __setitem__(self, name, value):
        """
        Set an entry in the local dictionary.

        Parameters
        ----------
        name : str
            name of the entry.
        value : -
            value of the entry to be value- and type-checked if declared.
        """
        if self._read_only:
            msg = "Tried to set '{}' on a read-only OptionsDictionary."
            raise KeyError(msg.format(name))

        # The key must have been declared.
        if name not in self._dict:
            msg = "Key '{}' cannot be set because it has not been declared."
            raise KeyError(msg.format(name))

        self._assert_valid(name, value)
        self._dict[name]['value'] = value

    def __getitem__(self, name):
        """
        Get an entry from the local dict, global dict, or declared default.

        Parameters
        ----------
        name : str
            name of the entry.

        Returns
        -------
        value : -
            value of the entry.
        """
        # If the entry has been set in this system, return the set value
        try:
            meta = self._dict[name]
            return meta['value']
        except KeyError:
            raise KeyError("Entry '{}' cannot be found".format(name))


class VectorAndNumbers:
    @classmethod
    def gen_number(cls, bounds=None, precision=0, distribution="uniform", p_type="real"):

        number = 0
        if bounds is None:
            bounds = [0, 1]

        if precision == 0:
            precision = 1e-12

        if distribution == "uniform":
            number = random() * (bounds[1] - bounds[0]) + bounds[0]
            number = round(number / precision) * precision

        if distribution == "normal":
            mean = (bounds[0] + bounds[1]) / 2
            std = (bounds[1] - bounds[0]) / 6
            number = normal(mean, std)

        if p_type == "integer":
            number = int(number)

        return number

    @classmethod
    def gen_vector(cls, design_parameters: dict):

        parameters_vector = []
        for parameter in design_parameters:

            if not ('bounds' in parameter):
                bounds = [parameter["initial_value"] * 0.5, parameter["initial_value"] * 1.5]
            else:
                bounds = parameter['bounds']

            if not ('precision' in parameter):
                precision = None
            else:
                precision = parameter['precision']

            if not ('parameter_type' in parameter):
                p_type = "real"
            else:
                p_type = parameter['parameter_type']

            if (precision is None) and (bounds is None):
                parameters_vector.append(cls.gen_number(p_type=p_type))
                continue

            if precision is None:
                parameters_vector.append(cls.gen_number(bounds=bounds, p_type=p_type))
                continue

            if bounds is None:
                parameters_vector.append(cls.gen_number(precision=precision, p_type=p_type))
                continue

            parameters_vector.append(cls.gen_number(bounds, precision))

        return parameters_vector.copy()



import os
import unittest
import tempfile
import time
from sqlitedict import SqliteDict

from artap.problem import Problem, ProblemViewDataStore
from artap.datastore import FileDataStore
from artap.algorithm_scipy import ScipyOpt
from artap.algorithm_sweep import SweepAlgorithm

from artap.results import Results
from artap.operators import RandomGenerator, CustomGenerator

from sys import platform
if platform == "win32":
    __platform__ = 'WINDOWS'
else:
    __platform__ = 'Linux'


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "NLopt_BOBYQA"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        x_1 = individual.vector[0]
        x_2 = individual.vector[1]

        # set custom properties
        individual.custom["functions"] = [x_1**2, x_2**2]

        return [x_1**2 + x_2**2]


class TestDataStoreFile(unittest.TestCase):
    def test_read_write_database(self):
        problem = MyProblem()

        # set data store
        database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".sqlite").name
        problem.data_store = FileDataStore(problem, database_name=database_name)

        algorithm = ScipyOpt(problem)
        algorithm.options['algorithm'] = 'CG'
        algorithm.options['tol'] = 1e-8
        algorithm.options['verbose_level'] = 0
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F')
        self.assertAlmostEqual(optimum.costs[0], 0, 3)

        # remove datastore
        problem.data_store.destroy()

        # check db
        db = SqliteDict(database_name, autocommit=True)

        populations = db["populations"]

        individual = populations[-1].individuals[9]
        self.assertAlmostEqual(individual.costs[0], 0, 3) # result
        self.assertAlmostEqual(individual.custom["functions"][0], individual.vector[0]**2)
        db.close()

        # remove file
        # print(database_name)
        os.remove(database_name)

    def xtest_read_datastore(self):
        problem = MyProblem()

        database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".sqlite").name
        problem.data_store = FileDataStore(problem, database_name=database_name, mode=FileMode.REWRITE)

        gen = CustomGenerator(problem.parameters)
        gen.init([[1, 2], [3, 3]])

        algorithm = SweepAlgorithm(problem, generator=gen)
        algorithm.run()

        problem.data_store.destroy()

    def test_read_datastore(self):
        database_name = "data" + os.sep + "data.sqlite"
        problem = ProblemViewDataStore(database_name=database_name)

        self.assertEqual(problem.name, 'NLopt_BOBYQA')

        individuals = problem.populations[-1].individuals
        self.assertEqual(individuals[0].vector[0], 1)
        self.assertEqual(individuals[1].vector[1], 3)
        self.assertEqual(individuals[0].costs, [5])
        self.assertEqual(individuals[1].costs, [18])


class TestDataStoreFileBenchmark(unittest.TestCase):
    def setUp(self):
        self.n = 20000

    def test_benchmark_data_store(self):
        t_s = time.time()
        problem = MyProblem()

        # set data store
        database_name = tempfile.NamedTemporaryFile(mode="w", delete=False, dir=None, suffix=".sqlite").name
        problem.data_store = FileDataStore(problem, database_name=database_name)

        gen = RandomGenerator(problem.parameters)
        gen.init(self.n)

        algorithm = SweepAlgorithm(problem, generator=gen)
        algorithm.options['max_processes'] = 1
        algorithm.run()

        cost = problem.populations[0].individuals[int(self.n / 2)].costs[0]

        # sync
        problem.data_store.destroy()

        t = time.time() - t_s
        problem.logger.info("write elapsed time: {} s, size: {} MB, n: {}".format(t, self.get_size(database_name) / 1024 / 1024, self.n))

        # check db
        t_s = time.time()
        db = SqliteDict(database_name, autocommit=True)

        populations = db["populations"]

        self.assertEqual(len(populations[0].individuals), self.n)
        self.assertAlmostEqual(populations[0].individuals[int(self.n / 2)].costs[0], cost, 3)
        db.close()

        # remove file
        # print(database_name)
        os.remove(database_name)

        t = time.time() - t_s
        problem.logger.info("read elapsed time: {} s".format(t))

    def get_size(self, start_path):
        if os.path.isfile(start_path):
            return os.path.getsize(start_path)
        else:
            total_size = 0
            for dirpath, dirnames, filenames in os.walk(start_path):
                for f in filenames:
                    fp = os.path.join(dirpath, f)
                    # skip if it is symbolic link
                    if not os.path.islink(fp):
                        total_size += os.path.getsize(fp)

            return total_size


if __name__ == '__main__':
    unittest.main()



from artap.problem import Problem
from artap.algorithm_genetic import NSGAII
from artap.results import Results

import matplotlib.pyplot as plt


class BiObjectiveTestProblem(Problem):
    """
    The goal of this example to show, how we can use Artap to solve a simple,
    bi-objective optimization problem.

    The problem is defined in the following way [GDE3]:

    Minimize f1 = x1
    Minimize f2 = (1+x2) / x1

    subject to
            x1 e [0.1, 1]
            x2 e [0, 5]

    The Pareto - front of the following problem is known, it is a simple
    hyperbola. This problem is very simple for an Evolutionary algorithm, it finds its solution within 20-30 generations.
    NSGA - II algorithm is used to solve this example.

    References
    ----------

    .: [GDE3] The third Evolution Step of Generalized Differential Evolution
    Saku Kukkonen, Jouni Lampinen

    """

    def set(self):
        # Not mandatory to give a name for the test problem
        self.name = 'Biobjective Test Problem'
        self.working_dir = '.'
        # Defines x_1 and x_2, which are the optimized parameters
        # and the bounds 'defines' the constraints of the optimization problem
        # nsga -- ii algorithm doesn't need an initial value for the definition
        self.parameters = [{'name': 'x_1', 'bounds': [0.1, 1.]},
                           {'name': 'x_2', 'bounds': [0.0, 5.0]}]

        # The two, separate optimization functions and the direction of the optimization
        # is set to minimization. It is also possible to use the maximize keyword.
        self.costs = [{'name': 'f_1', 'criteria': 'minimize'},
                      {'name': 'f_2', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        # The individual.vector function contains the problem parameters in the appropriate (previously defined) order
        f1 = individual.vector[0]
        f2 = (1 + individual.vector[1]) / individual.vector[0]
        # individual.auxvar = [1.]
        return [f1, f2]


# Initialization of the problem
problem = BiObjectiveTestProblem()

# Perform the optimization iterating over 100 times on 100 individuals.
algorithm = NSGAII(problem)
algorithm.options['max_population_number'] = 100
algorithm.options['max_population_size'] = 100
algorithm.run()

# Post - processing the results
# reads in the result values into the b, results class
b = Results(problem)
# print(b.parameters())
# b.write_out_populations()
solution = b.pareto_values()
# Plotting out the resulting hyperbola with matplotlib
plt.scatter([s[0] for s in solution],
            [s[1] for s in solution])

plt.xlim([0, 1.05])
plt.ylim([0, 10.1])

plt.xlabel("$f_1(x)$")
plt.ylabel("$f_2(x)$")

plt.show()



"""
Example shows how to use Artap together with localy installed Comsol Multiphysics

"""

from artap.executor import LocalComsolExecutor
from artap.problem import Problem
from artap.algorithm_scipy import ScipyOpt


class ComsolProblem(Problem):
    """
    Describe simple one objective optimization problem.

    """

    def set(self):
        self.name = "ComsolProblem"

        # Parameters must be defined in the Comsol model
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0.1, 100]},  # parameters a, b must be positive
                           {'name': 'b', 'initial_value': 10, 'bounds': [0.1, 100]}]

        self.costs = [{'name': 'F1', 'criteria': 'minimize'}]
        self.output_files = ["energy.txt"]

        # Executor serves for calling the Comsol Multiphisics
        self.executor = LocalComsolExecutor(self,
                                            problem_file="capacitor/capacitor.mph",  # File with the model
                                            output_files=self.output_files)  # file with results produced by Comsol

    # Calculate the value of the objective function
    def evaluate(self, individual):
        U = 10
        C_req = 5e-12  # the capacity 5 pF is required
        energy = self.executor.eval(individual)[0]
        capacity = 2 * energy / U ** 2
        f = abs(capacity - C_req)
        print(capacity)
        return [f / C_req]  # method evaluate must return list

    # This method processes files generated by 3rd party software, depends on files format
    def parse_results(self, output_files, individual):
        output_file = output_files[0]
        path = output_file
        with open(path) as file:
            content = file.read()

        lines = content.split("\n")
        line_with_results = lines[5]  # 5th line contains results

        result = float(line_with_results.split()[2])
        return [result]


if __name__ == "__main__":
    problem = ComsolProblem()  # Creating problem
    algorithm = ScipyOpt(problem)  # Algorithm from Scipy was choosen
    algorithm.options['algorithm'] = 'COBYLA'
    algorithm.options['n_iterations'] = 5
    algorithm.run()



"""
This example shows how to use Artap to solve a simple mathematical, single objective optimization problem.
The understanding of this problem doesn't need any physical knowledge.

Problem:
--------

A spyder, S, sits in one corner of a cuboid room, measuring 'H' by 'W' by 'L', and a fly, F, sits in the opposite
corner. By travelling on the surfaces of the room the, what is the shortest “straight line” distance from S to F?

The problem formulated as a single optimization problem, free optimization problem (without constraints)

Problem parameters are the edges of the cuboid: H, W, L
"""

from artap.problem import Problem
from artap.results import Results
from artap.algorithm_scipy import ScipyOpt

# wall lengths defined as global parameters
H = 1.
W = 1.
L = 1.


class ArtapProblem(Problem):
    """
    The solution of this problem needs to find the minimum of a one parametered (x_1) goal function.
    The problem solved, by the Nelder-Mead method, therefore an initial value has to be defined, anyway it is set to 0.
    """

    def set(self):

        self.name = "spyder on the wall"
        self.parameters = [{'name': 'x_1', 'initial_value':0.1, 'bounds': [0.0, 0.9]}]
        self.costs = [{'name': 'F_1'}]

    def evaluate(self, individual):
        # this function should be rewritten to define the optimization task
        # every individual contains a vector of parameters, which contains the previously defined problem parameters
        # in the given order
        x = individual.vector

        # the goal function
        F1 = (x[0] ** 2. + H ** 2.) ** 0.5 + ((W - x[0]) ** 2. + L ** 2.) ** 0.5

        # the evaluate function should give back a list of the calculated objective values, following the defined
        # order in set(Problem) in this case --> ['F1']
        return [F1]

### Optimization with Nelder-Mead
problem_nlm = ArtapProblem()

# set the optimization method
algorithm_nlm = ScipyOpt(problem_nlm)
algorithm_nlm.options['algorithm'] = 'Nelder-Mead'
algorithm_nlm.options['tol'] = 1e-3
algorithm_nlm.options['calculate_gradients'] = False

# perform the optimization
algorithm_nlm.run()

results_nlm = Results(problem_nlm)
opt = results_nlm.find_minimum('F_1')

print('The exact value of the optimization is at x_1 = 0.5')
print('Optimal solution (Nelder-Mead):', opt)



import unittest
from artap.problem import Problem
from artap.individual import Individual
from artap.job import Job


class JobProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "JobProblem"
        self.parameters = {'x_1': {'initial_value': 10}}
        self.costs = ['F_1']

    def evaluate(self, individual):
        result = 0
        for i in individual.vector:
            result += i*i

        return [result]

    def evaluate_constraints(self, individual):
        pass


class TestJob(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_individual_evaluation(self):
        problem = JobProblem()
        individual = Individual([1, 2, 2])
        job = Job(problem)
        job.evaluate(individual)
        self.assertEqual(individual.costs, [9])


if __name__ == '__main__':
    unittest.main()



import unittest
from artap.problem import Problem
from artap.algorithm_swarm import PSO, PSO_V1
from artap.results import Results
from artap.benchmark_functions import BinhAndKorn, AlpineFunction, Ackley


class TestAckley(unittest.TestCase):
    """ Tests that the PSO can find the global optimum. """

    def test_local_problem(self, population_number=30):
        try:
            problem = Ackley(**{'dimension': 1})
            algorithm = PSO(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 30
            algorithm.options['max_processes'] = 10
            algorithm.run()

            b = Results(problem)
            optimum = b.find_minimum('F_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))

class TestAckleyv1(unittest.TestCase):
    """ Tests that the PSO can find the global optimum. """

    def test_local_problem(self, population_number=30):
        try:
            problem = Ackley(**{'dimension': 1})
            algorithm = PSO_V1(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 30
            algorithm.options['max_processes'] = 10
            algorithm.run()

            b = Results(problem)
            optimum = b.find_minimum('F_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


# Very poor performance

# class TestAlpine(unittest.TestCase):
#
#
#     def test_local_problem(self, population_number=30):
#         try:
#             problem = AlpineFunction(**{'dimension': 3})
#             algorithm = PSO(problem)
#             algorithm.options['max_population_number'] = population_number
#             algorithm.options['max_population_size'] = 100
#             algorithm.options['max_processes'] = 3
#             algorithm.run()
#
#             b = Results(problem)
#             optimum = b.find_minimum('F_1')  # Takes last cost function
#             self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
#         except AssertionError:
#             # stochastic
#             print("TestAlpine::test_local_problem", population_number)
#             self.test_local_problem(int(1.5 * population_number))


# class TestAlpinev1(unittest.TestCase):
#
#
#     def test_local_problem(self, population_number=30):
#         try:
#             problem = AlpineFunction(**{'dimension': 3})
#             algorithm = PSO(problem)
#             algorithm.options['max_population_number'] = population_number
#             algorithm.options['max_population_size'] = 100
#             algorithm.options['max_processes'] = 3
#             algorithm.run()
#
#             b = Results(problem)
#             optimum = b.find_minimum('F_1')  # Takes last cost function
#             self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 1)
#         except AssertionError:
#             # stochastic
#             print("TestAlpine::test_local_problem", population_number)
#             self.test_local_problem(int(1.5 * population_number))
#


if __name__ == '__main__':
    unittest.main()



import unittest
from artap.individual import Individual
from artap.benchmark_robust import Synthetic1D, Synthetic2D, Synthetic5D, Synthetic10D


class TestSynthetic1D(unittest.TestCase):

    def test_synthetic1d(self):
        test = Synthetic1D()
        self.assertAlmostEqual(test.evaluate(Individual([11.0]))[0], 3.23, 3)
        self.assertAlmostEqual(test.evaluate(Individual([1.6]))[0], 3.205, 2)


class TestSynthetic2D(unittest.TestCase):

    def test_synthetic2d(self):
        test = Synthetic2D()
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 4.0]))[0], 1.21112, 4)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 1.0]))[0], 1.00096, 4)


class TestSynthetic5D(unittest.TestCase):

    def test_synthetic5d(self):
        test = Synthetic5D()
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 4.0, 1.3, 5.0, 5.0]))[0], 1.200000000, 4)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 1.0, 3.0, 2.0, 5.0]))[0], 1.000, 4)

        self.assertAlmostEqual(test.evaluate(Individual([10., 1.0, 6.0, 7.0, 8.0]))[0], .7)
        self.assertAlmostEqual(test.evaluate(Individual([1.0, 3.0, 8.0, 9.5, 2.0]))[0], .75)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 1.0, 3.0, 2.0, 5.0]))[0], 1.0)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 4.0, 1.3, 5.0, 5.0]))[0], 1.2, 5)

        self.assertAlmostEqual(test.evaluate(Individual([5.0, 2.0, 9.6, 7.3, 8.6]))[0], 1.0)
        self.assertAlmostEqual(test.evaluate(Individual([7.5, 8.0, 9.0, 3.2, 4.6]))[0], .6, 4)
        self.assertAlmostEqual(test.evaluate(Individual([5.7, 9.3, 2.2, 8.4, 7.1]))[0], .5)
        self.assertAlmostEqual(test.evaluate(Individual([5.5, 7.2, 5.8, 2.3, 4.5]))[0], .2, 4)

        self.assertAlmostEqual(test.evaluate(Individual([4.7, 3.2, 5.5, 7.1, 3.3]))[0], 0.4)
        self.assertAlmostEqual(test.evaluate(Individual([9.7, 8.4, 0.6, 3.2, 8.5]))[0], 0.1)


class TestSynthetic10D(unittest.TestCase):

    def test_synthetic10d(self):
        test = Synthetic10D()
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 4.0, 1.3, 5.0, 5.0, 3.0, 4.0, 1.3, 5.0, 5.0]))[0],
                               1.200000000, 4)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 1.0, 3.0, 2.0, 5.0, 3.0, 1.0, 3.0, 2.0, 5.0]))[0], 1.000,
                               4)
        self.assertAlmostEqual(test.evaluate(Individual([10., 1.0, 6.0, 7.0, 8.0, 1.0, 1.0, 6.0, 7.0, 8.0]))[0], 0.7)
        self.assertAlmostEqual(test.evaluate(Individual([1.0, 3.0, 8.0, 9.5, 2.0, 1.0, 3.0, 8.0, 9.5, 2.0]))[0], 0.75)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 1.0, 3.0, 2.0, 5.0, 3.0, 1.0, 3.0, 2.0, 5.0]))[0], 1.0)
        self.assertAlmostEqual(test.evaluate(Individual([3.0, 4.0, 1.3, 5.0, 5.0, 3.0, 4.0, 1.3, 5.0, 5.0]))[0], 1.2)



"""
algorithm.py
===========================================
This is base class for all algorithms
"""

from .problem import Problem
from .utils import ConfigDictionary
from abc import ABCMeta
from .operators import Evaluator


class Algorithm(metaclass=ABCMeta):
    """ Base class for optimization algorithms. """

    def __init__(self, problem: Problem, name="Algorithm"):
        self.name = name
        self.problem = problem
        self.evaluator = Evaluator(self)
        self.parameters = problem.parameters
        self.options = ConfigDictionary()

        self.options.declare(name='verbose_level', default=1, lower=0,
                             desc='Verbose level')
        self.options.declare(name='calculate_gradients', default=False,
                             desc='Enable calculating of gradients')
        # max(int(2 / 3 * cpu_count(), 1)
        self.options.declare(name='max_processes', default=1,
                             desc='Max running processes')

        self.options.declare(name='n_iterations', default=10,
                             desc='Max number of iterations')

    def evaluate(self, individuals):
        self.evaluator.evaluate(individuals)


class DummyAlgorithm(Algorithm):
    """
    Dummy class for testing
    """

    def __init__(self, problem, name='DummyAlgorithm'):
        super().__init__(problem, name)

    def run(self):
        pass




import unittest
from artap.individual import Individual
from artap.benchmark_pareto import BiObjectiveTestProblem, DTLZI,DTLZII, DTLZIII, DTLZIV

class TestBiobjective(unittest.TestCase):

    def test_biobjective(self):
        test2d = BiObjectiveTestProblem()
        self.assertAlmostEqual(test2d.evaluate(Individual([1.0, 1.0]))[1], 2.)


class TestDTLZI(unittest.TestCase):

    def test_dtlzI(self):
        test2d = DTLZI(**{'dimension': 8, 'm':3})

        # The optimal values of this function are on a hyperplane
        x = [0.5, 0.5, 0.5, 0.5,0.5, 0.5, 0.5, 0.5]
        f0 = test2d.evaluate(Individual(x))[0]
        f1 = test2d.evaluate(Individual(x))[1]
        f2 = test2d.evaluate(Individual(x))[2]

        self.assertAlmostEqual(f0+f1+f2, 0.5)


class TestDTLZII(unittest.TestCase):

    def test_dtlzII(self):
        test2d = DTLZII(**{'dimension': 12, 'm':3})

        # The optimal values of this function are on a hyperplane

        x = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
        f0 = test2d.evaluate(Individual(x))[0]
        f1 = test2d.evaluate(Individual(x))[1]
        f2 = test2d.evaluate(Individual(x))[2]

        self.assertAlmostEqual(f0**2.0+f1**2.0+f2**2.0, 1.0)



class TestDTLZIII(unittest.TestCase):

    def test_dtlzIII(self):
        test2d = DTLZIII(**{'dimension': 12, 'm':3})

        # The optimal values of this function are on a hyperplane

        x = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
        f0 = test2d.evaluate(Individual(x))[0]
        f1 = test2d.evaluate(Individual(x))[1]
        f2 = test2d.evaluate(Individual(x))[2]

        self.assertAlmostEqual(f0**2.0+f1**2.0+f2**2.0, 1.0)


class TestDTLZIV(unittest.TestCase):

    def test_dtlzIV(self):
        test2d = DTLZIV(**{'dimension': 12, 'm':3})

        # The optimal values of this function are on a hyperplane

        x = [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
        f0 = test2d.evaluate(Individual(x))[0]
        f1 = test2d.evaluate(Individual(x))[1]
        f2 = test2d.evaluate(Individual(x))[2]

        self.assertAlmostEqual(f0**2.0+f1**2.0+f2**2.0, 1.0)




import os
import rpyc
from rpyc.core.service import ClassicService

from .individual import Individual

MONITOR_PORT = 16000


class ArtapMonitorService(ClassicService):
    def __init__(self, problem):
        super().__init__()
        self.problem = problem

        # self.artap_dir = "/var/lib/artap/"

    def on_connect(self, conn):
        super().on_connect(conn)
        # print("Connected (ArtapMonitorService): ", self, conn)

    def on_disconnect(self, conn):
        super().on_disconnect(conn)
        # print("Disconnected (ArtapMonitorService): ", self, conn)

    def stats(self):
        st = {}
        st["populations_count"] = len(self.problem.populations)
        st["individuals_count"] = 0
        st["individuals_stats"] = { "empty": 0, "in_progress": 0, "evaluated": 0, "failed": 0 }
        for population in self.problem.populations:
            st["individuals_count"] += len(population.individuals)
            for individual in population.individuals:
                if individual.state == Individual.State.EMPTY:
                    st["individuals_stats"]["empty"] += 1
                elif individual.state == Individual.State.IN_PROGRESS:
                    st["individuals_stats"]["in_progress"] += 1
                elif individual.state == Individual.State.EVALUATED:
                    st["individuals_stats"]["evaluated"] += 1
                elif individual.state == Individual.State.FAILED:
                    st["individuals_stats"]["failed"] += 1

        return st

    def populations(self):
        return self.problem.populations

    def database_name(self):
        if self.problem.data_store is not None:
            if os.path.exists(self.problem.data_store.database_name):
                self.problem.data_store.sync()
                return os.path.abspath(self.problem.data_store.database_name)
        else:
            return None

    def problem(self):
        return self.problem


class MonitorService:
    def __init__(self, problem):
        self.problem = problem
        self.server = None
        self.port = 0

        self.start_server()

    def start_server(self):
        port = MONITOR_PORT
        while True:
            try:
                self.server = rpyc.utils.server.ThreadedServer(ArtapMonitorService(self.problem), port=port, auto_register=False)
                break
            except OSError as e:
                # print("Error: {}, port = {}".format(e.strerror, port))
                if e.strerror == "Address already in use":
                    port = port + 1
                    continue

        self.port = self.server.port
        self.problem.logger.info("Problem monitor created at {} port {}".format(self.server.host, self.port))
        self.server.logger.quiet = False
        self.server._start_in_thread()




import unittest
import importlib.util
import os


def make_test_function(name, filename):
    def test(self):
        print(name)
        spec = importlib.util.spec_from_file_location(name, filename)
        script = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(script)
    return test


class TestExamples(unittest.TestCase):
    """ Tests all examples in examples directory."""
    @staticmethod
    def init_tests():
        path = os.getcwd() + '/../examples/'
        for module in os.listdir(path):
            if module == '__init__.py' or module[-3:] != '.py':
                continue

            name = module[:-3]
            filename = path + os.sep + module
            test_func = make_test_function(name, filename)
            setattr(TestExamples, 'test_{0}'.format(name), test_func)

    # def test_all_examples(self):
    #
    #     path = os.getcwd() + '/../examples/'
    #     for module in os.listdir(path):
    #         if module == '__init__.py' or module[-3:] != '.py':
    #             continue
    #         spec = importlib.util.spec_from_file_location(module[:-3], path + os.sep + module)
    #         script = importlib.util.module_from_spec(spec)
    #         spec.loader.exec_module(script)
    #     del module


# create tests dynamically
TestExamples.init_tests()

if __name__ == '__main__':
    unittest.main()



from numpy.random import random_sample
from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm, GeneticIndividual
from .operators import FireflyMutator, DummySelector, RandomGenerator

import time


class MoFirefly(GeneticAlgorithm):
    """
    Multi-objective firefly algorithm
    """

    def __init__(self, problem: Problem, name="Particle Swarm Algorithm"):
        super().__init__(problem, name)
        self.n = self.options['max_population_size']
        self.mutator = FireflyMutator(self.problem.parameters)
        self.selector = DummySelector(self.problem.parameters, self.problem.signs)

    def run(self):
        self.generator = RandomGenerator(self.problem.parameters, individual_class=GeneticIndividual)
        self.generator.init(self.options['max_population_size'])

        population = self.gen_initial_population()
        self.selector.sorting(population.individuals)

        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))

        nr_gen = 0
        non_dominated_sol = []
        while nr_gen < self.options['max_population_number']:
            offsprings = self.selector.select(population.individuals)
            for i, a in enumerate(offsprings):
                non_dominated = True
                for j, b in enumerate(offsprings):
                    if i != j:
                        if self.mutator.dominate(a, b):
                            # in this case the individual a moves one in the direction b
                            # the step size is calculated from the intensity and the attraction
                            self.mutator.mutate_ij(a, b)
                            non_dominated = False
                            # self.evaluate([a]) <- if its not true a new one should be generated, this should be
                            #                      managed by the parallelization of the evaluator
                if non_dominated:
                    non_dominated_sol.append(a) # collects the non-dominated solutions for an iteration

            self.evaluate(offsprings)

            # to improve the convergence one solution from the pareto front is selected as global best
            # generating random weights for the cost function to select the 'global best'
            weights = random_sample(len(self.problem.costs))
            c = 1. / sum(weights)
            weights = [x * c for x in weights]

            self.problem.populations[-1] = offsprings

            nr_gen += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))



from random import randint
from .problem import Problem
from .population import Population
from .algorithm_genetic import GeneticAlgorithm, GeneticIndividual
from .operators import SwarmMutator, DummySelector, RandomGenerator, SwarmMutatorTVIW

import time


class PSO(GeneticAlgorithm):

    def __init__(self, problem: Problem, name="Particle Swarm Algorithm"):
        super().__init__(problem, name)
        self.n = self.options['max_population_size']
        self.mutator = SwarmMutator(self.problem.parameters)
        self.selector = DummySelector(self.problem.parameters, self.problem.signs)

    def run(self):
        # set random generator
        self.generator = RandomGenerator(self.problem.parameters, individual_class=GeneticIndividual)
        self.generator.init(self.options['max_population_size'])

        population = self.gen_initial_population()

        for individual in population.individuals:
            self.mutator.evaluate_best_individual(individual)

        self.selector.sorting(population.individuals)

        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))

        i = 0
        while i < self.options['max_population_number']:
            offsprings = self.selector.select(population.individuals)

            pareto_front = []
            for individual in offsprings:
                if individual.front_number == 1:
                    pareto_front.append(individual)

            for individual in offsprings:
                index = randint(0, len(pareto_front) - 1)  # takes random individual from Pareto front
                best_individual = pareto_front[index]
                if best_individual is not individual:
                    self.mutator.update(best_individual)
                    self.mutator.mutate(individual)

            population = Population(offsprings)
            self.problem.populations.append(population)
            self.evaluator.evaluate(offsprings)

            for individual in offsprings:
                self.mutator.evaluate_best_individual(individual)

            self.selector.sorting(offsprings)



            i += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))


class PSO_V1(GeneticAlgorithm):
    """
    This algorithm is a variant of the original PSO, published by Eberhart(2000), the origin of this modified algorithm,
    which constriction factor was introduced by Clercs in 1999.

    The code is based on SHI and EBERHARTS algorithm.

    Empirical study of particle swarm optimization,” in Proc. IEEE Int. Congr. Evolutionary Computation, vol. 3,
    1999, pp. 101–106.
    """

    def __init__(self, problem: Problem, name="Particle Swarm Algorithm - with time varieting inertia weight"):
        super().__init__(problem, name)
        self.n = self.options['max_population_size']
        self.mutator = SwarmMutatorTVIW(self.problem.parameters, self.options['max_population_number'])
        self.selector = DummySelector(self.problem.parameters, self.problem.signs)

    def run(self):
        # set random generator
        self.generator = RandomGenerator(self.problem.parameters, individual_class=GeneticIndividual)
        self.generator.init(self.options['max_population_size'])

        population = self.gen_initial_population()

        for individual in population.individuals:
            self.mutator.evaluate_best_individual(individual)

        self.selector.sorting(population.individuals)
        self.problem.populations.append(population)

        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))

        i = 0
        while i < self.options['max_population_number']:
            offsprings = self.selector.select(population.individuals)

            pareto_front = []
            for individual in offsprings:
                if individual.front_number == 1:
                    pareto_front.append(individual)

            for individual in offsprings:
                index = randint(0, len(pareto_front) - 1)  # takes random individual from Pareto front
                best_individual = pareto_front[index]
                if best_individual is not individual:
                    self.mutator.update(best_individual)
                    self.mutator.mutate(individual)

            self.evaluate(offsprings)

            for individual in offsprings:
                self.mutator.evaluate_best_individual(individual)

            self.selector.sorting(offsprings)
            population = Population(offsprings)
            self.problem.populations.append(population)

            i += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))


import unittest

from artap.problem import Problem
from artap.benchmark_functions import BinhAndKorn, Ackley
from artap.algorithm_genetic import EpsMOEA
from artap.results import Results, GraphicalResults

# import optproblems as optp
# import optproblems.cec2005 as cec2005

# class BinhAndKornProblem(Problem):
#
#     """ Describe simple one objective optimization problem. """
#     def set(self):
#         self.name = "EPSMOEA"
#         self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5]},
#                            {'name': 'x_2', 'initial_value': 1.5, 'bounds': [0, 3]}]
#         self.costs = [{'name': 'F_1'}, {'name': 'F_2'}]
#
#     def evaluate(self, x):
#         function = BinhAndKorn()
#         return function.eval(x)
#
#     def evaluate_constraints(self, x):
#         return BinhAndKorn.constraints(x)
#
#
# class TestEPSMOEAOptimization(unittest.TestCase):
#     """ Tests simple one objective optimization problem."""
#
#     def test_local_problem_epsmoea(self):
#         problem = BinhAndKornProblem()
#         algorithm = EpsMOEA(problem)
#         algorithm.options['max_population_number'] = 100
#         algorithm.options['max_population_size'] = 100
#         algorithm.options['calculate_gradients'] = True
#         algorithm.options['verbose_level'] = 1
#         algorithm.options['epsilons'] = 0.05
#
#         algorithm.run()
#
#         b = Results(problem)
#         solution = b.pareto_values()
#         wrong = 0
#         for sol in solution:
#             if abs(BinhAndKorn.approx(sol[0]) - sol[1]) > 0.1 * BinhAndKorn.approx(sol[0]) \
#                     and 20 < sol[0] < 70:
#                 wrong += 1
#
#         self.assertLessEqual(wrong, 10)
#
#
# class AckleyN2Test(Problem):
#     """Test the convergence in a one objective example with a simple 2 variable Ackley N2 formula"""
#
#     def set(self):
#         self.name = "TestAckleyN2"
#         self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-32, 32]},
#                       {'name': 'x_2', 'initial_value': 2.5, 'bounds': [-32, 32]}]
#         self.costs = [{'name': 'F_1'}]
#
#     def evaluate(self, individual):
#         function = AckleyN2()
#         return [function.eval(individual.vector)]


class TestAckleyN222(unittest.TestCase):
    """ Tests that the eps-moea algorithm can find the global optimum of a function."""

    def test_local_problem(self, population_number=10):
        try:
            problem = Ackley(**{'dimension':1})
            algorithm = EpsMOEA(problem)
            algorithm.options['max_population_number'] = population_number
            algorithm.options['max_population_size'] = 100
            algorithm.options['epsilons'] = 0.01
            algorithm.options['max_processes'] = 10
            algorithm.run()

            b = Results(problem)
            optimum = b.find_minimum('F_1')  # Takes last cost function
            self.assertAlmostEqual(optimum.costs[0], problem.global_optimum, 0)
        except AssertionError:
            # stochastic
            print("TestAckleyN222::test_local_problem", population_number)
            self.test_local_problem(int(1.5 * population_number))


if __name__ == '__main__':
    unittest.main()



import unittest
import math

from artap.problem import Problem
from artap.individual import Individual
from artap.results import Results

try:
    from artap.algorithm_nlopt import NLopt
    from artap.algorithm_nlopt import LN_BOBYQA

    __nlopt__ = True
except ImportError:
    print("NLopt is not present test skiped")
    __nlopt__ = False

try:
    from agrossuite import agros as a2d

    __agros__ = True
except ImportError:
    print("Agros is not present test skipped")
    __agros__ = False


class AgrosProblem(Problem):

    def set(self):
        self.name = "agros problem"
        self.working_dir = "team_22/"
        self.parameters = [{'name': 'R2', 'initial_value': 3.0, 'bounds': [2.6, 3.4]},
                           {'name': 'h2', 'initial_value': 1.0, 'bounds': [0.408, 2.2]},
                           {'name': 'd2', 'initial_value': 0.3, 'bounds': [0.1, 0.4]}]

        self.costs = [{'name': 'F', 'criteria': 'minimize', 'weight': 0}]

    def evaluate(self, individual: Individual):
        # problem
        x = individual.vector
        problem = a2d.problem(clear=True)
        problem.coordinate_type = "axisymmetric"
        problem.mesh_type = "triangle"
        problem.parameters["J1"] = 2.25e+7
        problem.parameters["J2"] = -2.25e+7
        problem.parameters["R1"] = 2.0
        problem.parameters["R2"] = x[0]
        problem.parameters["d1"] = 0.27
        problem.parameters["d2"] = x[2]
        problem.parameters["h1"] = 1.6
        problem.parameters["h2"] = x[1]

        # fields
        # magnetic
        magnetic = problem.field("magnetic")
        magnetic.analysis_type = "steadystate"
        magnetic.matrix_solver = "umfpack"
        magnetic.number_of_refinements = 2
        magnetic.polynomial_order = 2
        magnetic.adaptivity_type = "disabled"
        magnetic.solver = "linear"

        # boundaries
        magnetic.add_boundary("A = 0", "magnetic_potential", {"magnetic_potential_real": 0})
        magnetic.add_boundary("Neumann", "magnetic_surface_current", {"magnetic_surface_current_real": 0})

        # materials
        magnetic.add_material("Coil 1", {"magnetic_conductivity": 0, "magnetic_current_density_external_real": "J1",
                                         "magnetic_permeability": 1, "magnetic_remanence": 0,
                                         "magnetic_remanence_angle": 0, "magnetic_total_current_prescribed": 0,
                                         "magnetic_total_current_real": 0, "magnetic_velocity_angular": 0,
                                         "magnetic_velocity_x": 0, "magnetic_velocity_y": 0})
        magnetic.add_material("Coil 2", {"magnetic_conductivity": 0, "magnetic_current_density_external_real": "J2",
                                         "magnetic_permeability": 1, "magnetic_remanence": 0,
                                         "magnetic_remanence_angle": 0, "magnetic_total_current_prescribed": 0,
                                         "magnetic_total_current_real": 0, "magnetic_velocity_angular": 0,
                                         "magnetic_velocity_x": 0, "magnetic_velocity_y": 0})
        magnetic.add_material("Air", {"magnetic_conductivity": 0, "magnetic_current_density_external_real": 0,
                                      "magnetic_permeability": 1, "magnetic_remanence": 0,
                                      "magnetic_remanence_angle": 0, "magnetic_total_current_prescribed": 0,
                                      "magnetic_total_current_real": 0, "magnetic_velocity_angular": 0,
                                      "magnetic_velocity_x": 0, "magnetic_velocity_y": 0})

        # geometry
        geometry = problem.geometry()
        geometry.add_edge("R1-d1/2", 0, "R1+d1/2", 0, boundaries={"magnetic": "Neumann"})
        geometry.add_edge("R1+d1/2", 0, "R1+d1/2", "h1/2")
        geometry.add_edge("R1+d1/2", "h1/2", "R1-d1/2", "h1/2")
        geometry.add_edge("R1-d1/2", "h1/2", "R1-d1/2", 0)
        geometry.add_edge("R2-d2/2", "h2/2", "R2+d2/2", "h2/2")
        geometry.add_edge("R2+d2/2", "h2/2", "R2+d2/2", 0, boundaries={"magnetic": "Neumann"})
        geometry.add_edge("R2-d2/2", "h2/2", "R2-d2/2", 0)
        geometry.add_edge("R2-d2/2", 0, "R2+d2/2", 0, boundaries={"magnetic": "Neumann"})
        geometry.add_edge(0, 0, "R1-d1/2", 0, boundaries={"magnetic": "Neumann"})
        geometry.add_edge("R1+d1/2", 0, "R2-d2/2", 0, boundaries={"magnetic": "Neumann"})
        geometry.add_edge(0, 25, 0, 0, boundaries={"magnetic": "A = 0"})
        geometry.add_edge(25, 0, "R2+d2/2", 0, boundaries={"magnetic": "Neumann"})
        geometry.add_edge(0, 25, 25, 25, boundaries={"magnetic": "A = 0"})
        geometry.add_edge(25, 25, 25, 0, boundaries={"magnetic": "A = 0"})

        geometry.add_label(12, 12, materials={"magnetic": "Air"})
        geometry.add_label("R1", "h1/4", materials={"magnetic": "Coil 1"})
        geometry.add_label("R2", "h2/4", materials={"magnetic": "Coil 2"})

        # recipes
        magnetic.add_recipe_volume_integral("Wm", "magnetic_energy", [], -1, -1)
        magnetic.add_recipe_local_value("B00", "magnetic_flux_density_real", "magnitude", 0, 10, -1, -1)
        magnetic.add_recipe_local_value("B01", "magnetic_flux_density_real", "magnitude", 1, 10, -1, -1)
        magnetic.add_recipe_local_value("B02", "magnetic_flux_density_real", "magnitude", 2, 10, -1, -1)
        magnetic.add_recipe_local_value("B03", "magnetic_flux_density_real", "magnitude", 3, 10, -1, -1)
        magnetic.add_recipe_local_value("B04", "magnetic_flux_density_real", "magnitude", 4, 10, -1, -1)
        magnetic.add_recipe_local_value("B05", "magnetic_flux_density_real", "magnitude", 5, 10, -1, -1)
        magnetic.add_recipe_local_value("B06", "magnetic_flux_density_real", "magnitude", 6, 10, -1, -1)
        magnetic.add_recipe_local_value("B07", "magnetic_flux_density_real", "magnitude", 7, 10, -1, -1)
        magnetic.add_recipe_local_value("B08", "magnetic_flux_density_real", "magnitude", 8, 10, -1, -1)
        magnetic.add_recipe_local_value("B09", "magnetic_flux_density_real", "magnitude", 9, 10, -1, -1)
        magnetic.add_recipe_local_value("B10", "magnetic_flux_density_real", "magnitude", 10, 10, -1, -1)
        magnetic.add_recipe_local_value("B11", "magnetic_flux_density_real", "magnitude", 10, 10, -1, -1)
        magnetic.add_recipe_local_value("B12", "magnetic_flux_density_real", "magnitude", 10, 9, -1, -1)
        magnetic.add_recipe_local_value("B13", "magnetic_flux_density_real", "magnitude", 10, 8, -1, -1)
        magnetic.add_recipe_local_value("B14", "magnetic_flux_density_real", "magnitude", 10, 7, -1, -1)
        magnetic.add_recipe_local_value("B15", "magnetic_flux_density_real", "magnitude", 10, 6, -1, -1)
        magnetic.add_recipe_local_value("B16", "magnetic_flux_density_real", "magnitude", 10, 5, -1, -1)
        magnetic.add_recipe_local_value("B17", "magnetic_flux_density_real", "magnitude", 10, 4, -1, -1)
        magnetic.add_recipe_local_value("B18", "magnetic_flux_density_real", "magnitude", 10, 3, -1, -1)
        magnetic.add_recipe_local_value("B19", "magnetic_flux_density_real", "magnitude", 10, 2, -1, -1)
        magnetic.add_recipe_local_value("B20", "magnetic_flux_density_real", "magnitude", 10, 1, -1, -1)
        magnetic.add_recipe_local_value("B21", "magnetic_flux_density_real", "magnitude", 10, 0, -1, -1)

        computation = problem.computation()
        computation.solve()

        solution = computation.solution("magnetic")

        B00 = solution.local_values(0, 10)["Br"]
        B01 = solution.local_values(1, 10)["Br"]
        B02 = solution.local_values(2, 10)["Br"]
        B03 = solution.local_values(3, 10)["Br"]
        B04 = solution.local_values(4, 10)["Br"]
        B05 = solution.local_values(5, 10)["Br"]
        B06 = solution.local_values(6, 10)["Br"]
        B07 = solution.local_values(7, 10)["Br"]
        B08 = solution.local_values(8, 10)["Br"]
        B09 = solution.local_values(9, 10)["Br"]
        B10 = solution.local_values(10, 10)["Br"]
        B11 = solution.local_values(10, 10)["Br"]
        B12 = solution.local_values(10, 9)["Br"]
        B13 = solution.local_values(10, 8)["Br"]
        B14 = solution.local_values(10, 7)["Br"]
        B15 = solution.local_values(10, 6)["Br"]
        B16 = solution.local_values(10, 5)["Br"]
        B17 = solution.local_values(10, 4)["Br"]
        B18 = solution.local_values(10, 3)["Br"]
        B19 = solution.local_values(10, 2)["Br"]
        B20 = solution.local_values(10, 1)["Br"]
        B21 = solution.local_values(10, 0)["Br"]

        Wm = solution.volume_integrals()["Wm"]

        of = (
                     B00 ** 2 + B01 ** 2 + B02 ** 2 + B03 ** 2 + B04 ** 2 + B05 ** 2 + B06 ** 2 + B07 ** 2 + B08 ** 2 + B09 ** 2 + B10 ** 2 + B11 ** 2 + B12 ** 2 + B13 ** 2 + B14 ** 2 + B15 ** 2 + B16 ** 2 + B17 ** 2 + B18 ** 2 + B19 ** 2 + B20 ** 2 + B21 ** 2) / 22 / 9e-6 + math.fabs(
            2 * Wm - 180e6) / 180e6

        return [of]


class TestAgrosOptimization(unittest.TestCase):

    @unittest.skipIf(__agros__ is False or __nlopt__ is False, "requires Agros Suite")
    def test_agros_exec(self):
        problem = AgrosProblem()
        algorithm = NLopt(problem)
        algorithm.options['n_iterations'] = 3
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['verbose_level'] = 0
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F')
        self.assertLessEqual(optimum.costs[0], 6.1)


if __name__ == '__main__':
    unittest.main()



from .problem import Problem
from .algorithm_genetic import GeneticAlgorithm
from .operators import GradientEvaluator, RandomGenerator
from .population import Population
from .individual import Individual


class GradientDescent(GeneticAlgorithm):
    """ Gradient descent algorithms """

    def __init__(self, problem: Problem, name="GradientDescent"):
        super().__init__(problem, name)
        self.problem = problem

        self.evaluator = GradientEvaluator(self)
        self.options.declare(name='n_iterations', default=10,
                             desc='Maximum number of iterations')

        self.options.declare(name='x0', default=0,
                             desc='initial point')

        self.options.declare(name='h', default=0.1,
                             desc='step')

    def run(self):
        h = self.options["h"]

        self.generator = RandomGenerator(self.problem.parameters)
        self.generator.init(self.options['max_population_size'])
        self.gen_initial_population()

        # TODO: add adaptive step size
        for it in range(self.options['max_population_number']):
            population = Population()
            for individual in self.problem.populations[it].individuals:
                x = []
                for i in range(len(individual.vector)):
                    x.append(individual.vector[i] - h * individual.features['gradient'][i])
                individual = Individual(x)
                population.individuals.append(individual)
            self.evaluate(population.individuals)
            self.problem.populations.append(population)




from artap.operators import SimpleMutator, SimulatedBinaryCrossover, SimpleCrossover, \
    TournamentSelector, ParetoDominance
from artap.individual import Individual
from artap.algorithm_genetic import GeneticIndividual

import unittest


class TestCrossover(unittest.TestCase):
    """ Tests crossover."""

    def setUp(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [0, 3]}]

        self.signs = [1, 1]

        self.i1 = Individual([1, 2, 2])
        self.i2 = Individual([3, 2, 1])

    def test_simple_mutation(self):
        sbx = SimpleMutator(self.parameters, 0.7)
        individual = sbx.mutate(self.i1)
        self.assertTrue(
            self.parameters[0]['bounds'][0] <= individual.vector[0] <= self.parameters[0]['bounds'][1] and
            self.parameters[1]['bounds'][0] <= individual.vector[1] <= self.parameters[1]['bounds'][1])

    def test_sbx(self):
        sbx = SimulatedBinaryCrossover(self.parameters, 10)
        offsprings = sbx.cross(self.i1, self.i2)
        self.assertEqual(len(offsprings), 2)

    def test_simple_crossover(self):
        sbx = SimpleCrossover(self.parameters, 0.9)
        offsprings = sbx.cross(self.i1, self.i2)
        self.assertEqual(len(offsprings), 2)

    def test_dominates(self):
        dominance = ParetoDominance()
        i1 = GeneticIndividual([0, 0])
        i1.costs = [1, 1]

        i1.transform_data(self.signs)

        i2 = GeneticIndividual([2, 2])
        i2.costs = [2, 2]

        i2.transform_data(self.signs)

        result = dominance.compare(i1.signed_costs, i2.signed_costs)
        self.assertEqual(result, 1)

        result = dominance.compare(i2.signed_costs, i1.signed_costs)
        self.assertEqual(result, 2)

    def test_pareto(self):
        individuals = []
        for i in range(3):
            for j in range(3):
                individual = GeneticIndividual([0, 1])
                individuals.append(individual)

        individuals[0].costs = [0, 1]
        individuals[1].costs = [1, 0]
        individuals[2].costs = [0, 2]
        individuals[3].costs = [1, 1]
        individuals[4].costs = [2, 0]
        individuals[5].costs = [0, 3]
        individuals[6].costs = [1, 2]
        individuals[7].costs = [2, 1]
        individuals[8].costs = [3, 0]

        _ = [ind.transform_data(self.signs) for ind in individuals]

        selector = TournamentSelector(self.parameters)
        selector.sorting(individuals)

        for individual in individuals:
            # print(individual.costs, individual.front_number)
            self.assertEqual(individual.costs[0] + individual.costs[1], individual.front_number)


if __name__ == '__main__':
    unittest.main()



from setuptools import Command, setup, find_packages
from setuptools.command.install import install
import os


class CustomInstallCommand(install):
    """Custom install setup to help run shell commands (outside shell) before installation"""

    def run(self):
        os.system('python3 -m pip install numpy')
        os.system('python3 -m pip install cython')
        install.run(self)


base_dir = os.path.abspath(os.path.dirname(__file__))

# long description
with open("README.md", "r") as fh:
    long_description = fh.read()

# requirements
with open("requirements.txt", "r") as fh:
    requirements = fh.read().split("\n")

min_numpy_ver = "1.13.3"

def setup_package():
    setuptools_kwargs = {
        "install_requires": [
            # "python-dateutil >= 2.6.1",
            # "pytz >= 2017.2",
            f"numpy >= {min_numpy_ver}",
            requirements,
            #"smt >= 0.3.4",
        ],
        "setup_requires": [f"numpy >= {min_numpy_ver}"],
        "zip_safe": False,
    }

    setup(
        name="artap",
        version="2020.4.02.1",
        author=u"Artap Team",
        author_email="artap.framework@gmail.com",
        description="Platform for robust design optimization",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="http://www.agros2d.org/artap/",
        python_requires='>=3.7',
        license="License :: OSI Approved :: MIT License",
        cmdclass={'install': CustomInstallCommand},
        packages=find_packages(),
        include_package_data=True,
        # data_files=[('artap/lib', ['artap/lib/bayesopt.so']),
        # install_requires=requirements,
        # scripts=['3rdparty/submodules.sh'],
        classifiers=[
            "Intended Audience :: Science/Research",
            "Operating System :: POSIX :: Linux",
            "Topic :: Scientific/Engineering",
            'Programming Language :: Python :: 3.7',
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        **setuptools_kwargs,
    )


if __name__ == "__main__":
    setup_package()



import unittest
from artap.problem import Problem
from artap.algorithm_gradient_descent import GradientDescent


class GradientProblem(Problem):
    """Test of calculation of gradientS"""

    def set(self):
        self.name = TestAckleyN2
        self.parameters = [{'name': 'x', 'initial_value': 2.13}, {'name': 'y', 'initial_value': 2.13}]
        self.costs = [{'name': 'F_1', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        x = individual.vector
        return [(x[0]-1)**2 + x[1]**2]


class TestAckleyN2(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem(self):
        problem = GradientProblem()
        algorithm = GradientDescent(problem)
        algorithm.options["x0"] = [10, 10]
        algorithm.options["n_iterations"] = 8
        algorithm.options["h"] = 0.1
        algorithm.run()


if __name__ == '__main__':
    unittest.main()



import unittest
# from artap.doe import build_box_behnken, build_lhs, build_frac_fact, build_full_fact, build_plackett_burman
# from artap.individual import Individual
from artap.operators import CustomGenerator, RandomGenerator, FullFactorGenerator, PlackettBurmanGenerator, \
    BoxBehnkenGenerator, LHSGenerator


class TestDOE(unittest.TestCase):

    def setUp(self):
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-2.5, 5]},
                           {'name': 'x_2', 'initial_value': 1.5, 'bounds': [1, 3.4]},
                           {'name': 'x_3', 'initial_value': 3.5, 'bounds': [6, 10]}]

    def test_random_generator(self):
        gen = RandomGenerator(parameters=self.parameters)
        gen.init(3)
        individuals = gen.generate()
        # values
        self.assertTrue(self.parameters[0]['bounds'][0] <= individuals[0].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[0].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[0].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[1].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[1].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[1].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[2].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[2].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[2].vector[2] <= self.parameters[2]['bounds'][1])

    def test_custom_generator(self):
        gen = CustomGenerator()
        gen.init([[3, 2, 6], [-1, 3, 8]])
        individuals = gen.generate()
        # values
        self.assertEqual(individuals[0].vector[0], 3)
        self.assertEqual(individuals[0].vector[1], 2)
        self.assertEqual(individuals[0].vector[2], 6)
        self.assertEqual(individuals[1].vector[0], -1)
        self.assertEqual(individuals[1].vector[1], 3)
        self.assertEqual(individuals[1].vector[2], 8)

    # Factorial Designs
    def test_full_fact_generator(self):
        # General Full-Factorial
        gen = FullFactorGenerator(parameters=self.parameters)
        gen.init(center=False)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 8)
        # values
        self.assertEqual(individuals[0].vector[0], -2.5)
        self.assertEqual(individuals[1].vector[0], 5.0)
        self.assertEqual(individuals[2].vector[1], 3.4)
        self.assertEqual(individuals[3].vector[1], 3.4)
        self.assertEqual(individuals[7].vector[2], 10.0)

        gen = FullFactorGenerator(parameters=self.parameters)
        gen.init(center=True)
        individuals = gen.generate()

        self.assertEqual(individuals[3].vector[0], -2.5)
        self.assertEqual(individuals[-1].vector[1], 3.4)

    def test_plackett_burman_generator(self):
        # Plackett-Burman
        gen = PlackettBurmanGenerator(parameters=self.parameters)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 4)
        # values
        self.assertEqual(individuals[0].vector[0], -2.5)
        self.assertEqual(individuals[1].vector[0], 5.0)
        self.assertEqual(individuals[2].vector[2], 6.0)
        self.assertEqual(individuals[3].vector[1], 3.4)

    # Response - Surface Designs
    def test_box_behnken_generator(self):
        # Box-Behnken
        gen = BoxBehnkenGenerator(parameters=self.parameters)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 13)
        # values
        self.assertEqual(individuals[2].vector[2], 8.0)
        self.assertEqual(individuals[3].vector[1], 3.4)
        self.assertEqual(individuals[5].vector[0], 5.0)
        self.assertEqual(individuals[12].vector[1], 2.2)

    # Randomized Designs
    def test_lhs_generation(self):
        # Latin - Hypercube
        gen = LHSGenerator(parameters=self.parameters)
        gen.init(number=3)
        individuals = gen.generate()

        # size
        self.assertEqual(len(individuals), 3)
        # values
        self.assertTrue(self.parameters[0]['bounds'][0] <= individuals[0].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[0].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[0].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[1].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[1].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[1].vector[2] <= self.parameters[2]['bounds'][1] and
                        self.parameters[0]['bounds'][0] <= individuals[2].vector[0] <= self.parameters[0]['bounds'][1] and
                        self.parameters[1]['bounds'][0] <= individuals[2].vector[1] <= self.parameters[1]['bounds'][1] and
                        self.parameters[2]['bounds'][0] <= individuals[2].vector[2] <= self.parameters[2]['bounds'][1])


if __name__ == '__main__':
    unittest.main()



"""
This function shows, how you can fit a function with NSGA II to your measured data.

The measured data is coming from a partial discharge measurement, where the initial slope of the voltage responses
measured. The task is to find out the R and C elements of the modeling R-C circuit. The theory says, that
these elements can be found if we can search these elements in the following, exponential shape:


The detailed description of the analytical model can be founc here: doi....
"""

import matplotlib.pyplot as plt
import numpy as np

# (tdchi, SRi)
SR2 = [(1.0, 100.7923), (2.0, 95.58673), (4.0, 85.98188), (6.0, 77.35857), (8.0, 69.6149), (10.0, 62.65971),
       (15.0, 48.20685), (20.0, 37.13707), (30.0, 22.12758), (50.0, 7.979435), (75.0, 2.29236), (100.0, 0.676797),
       (150.0, 0.062897), (200.0, 0.062897), (300.0, 6.49E-05), (500.0, 7.75E-09)]

x_val = [x[0] for x in SR2]
y_val = [x[1] for x in SR2]
plt.plot(x_val, y_val, 'bo', label="Measured data")

C2o1 = [9.627, 9.958385741447383, 9.68637611277249]
R2o1 = [5.870, 1.692170503816123, 2.3120386928356145]

# NSGA result - 200 generations
sol_nsga = [5.87, 9.627, 1.2254465803764465, 14.458027617643706, 4.8264795611077655, 5.208973499033958]
sol_qlnp = [R2o1[0], C2o1[0], R2o1[1], C2o1[1], R2o1[2], C2o1[2]]


def sri(x, tdch, Uch, tch, c0):
    """
    Calculates the ui voltage from the given parameters
    :param x: list of individuals from the optimized paramters in the ri, ci order
    :param Sri: measured data at the ith case, contains tuples : (tdchi, Sri) format
    :return:
    """
    sri = 0.
    for j in range(2, len(x), 2):  # step is 2, because r,c pairs
        taui = x[j] * x[j + 1]  # ri*ci
        sri += Uch * (1.0 - np.exp(-tch / taui)) * np.exp(-tdch / taui) / x[j] / c0
    return sri


Uch = 1000.  # V
tch = 4000.  # s
c0 = 9.627

y_nsga = [sri(sol_nsga, t[0], Uch, tch, c0) for t in SR2]
y_nm = [sri(sol_qlnp, t[0], Uch, tch, c0) for t in SR2]

plt.plot(x_val, y_nsga, 'blue', label="Fitted data -- NSGA --")
plt.plot(x_val, y_nm, 'r+--', label="Fitted data -- NLP --")

plt.legend(loc='upper left')
plt.xlabel(r"$\t$", fontsize=16)
plt.show()



import math
from artap.problem import Problem
from artap.individual import Individual
from artap.algorithm_genetic import NSGAII
from artap.algorithm_nlopt import NLopt, LN_BOBYQA
from artap.results import Results, GraphicalResults
import agrossuite.agros as a2d

from artap.examples.TEAM_BENCHMARK_35x.semi_analytical_optimization import ProblemAnalytical


class AgrosSimple(Problem):

    def set(self):

        self.name = "Agros-solution"

        self.parameters = [{'name': 'x1', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x2', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x3', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x4', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x5', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x6', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x7', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x8', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x9', 'bounds': [5.01e-3, 50e-3]},
                           {'name': 'x10', 'bounds': [5.01e-3, 50e-3]}]

        self.costs = [{'name': 'F1', 'criteria': 'minimize'},
                      {'name': 'F2', 'criteria': 'minimize'}]

    def evaluate(self, x):

        x = x.vector

        # problem
        problem = a2d.problem(clear=True)
        problem.coordinate_type = "axisymmetric"
        problem.mesh_type = "triangle"

        problem.parameters["MH"] = 0.04
        problem.parameters["MW"] = 0.1
        problem.parameters["h"] = 0.00145
        problem.parameters["r1"] = x[0]
        problem.parameters["r10"] = x[9]
        problem.parameters["r2"] = x[1]
        problem.parameters["r3"] = x[2]
        problem.parameters["r4"] = x[3]
        problem.parameters["r5"] = x[4]
        problem.parameters["r6"] = x[5]
        problem.parameters["r7"] = x[6]
        problem.parameters["r8"] = x[7]
        problem.parameters["r9"] = x[8]
        problem.parameters["w"] = 0.001

        # fields
        # magnetic
        magnetic = problem.field("magnetic")
        magnetic.analysis_type = "steadystate"
        magnetic.matrix_solver = "umfpack"
        magnetic.number_of_refinements = 1
        magnetic.polynomial_order = 2
        magnetic.adaptivity_type = "disabled"
        magnetic.solver = "linear"

        # boundaries
        magnetic.add_boundary("A=0", "magnetic_potential", {"magnetic_potential_real": 0})
        magnetic.add_boundary("Symmetry", "magnetic_surface_current", {"magnetic_surface_current_real": 0})

        # materials
        magnetic.add_material("air", {"magnetic_conductivity": 0, "magnetic_current_density_external_real": 0,
                                      "magnetic_permeability": 1, "magnetic_remanence": 0,
                                      "magnetic_remanence_angle": 0, "magnetic_total_current_prescribed": 0,
                                      "magnetic_total_current_real": 0, "magnetic_velocity_angular": 0,
                                      "magnetic_velocity_x": 0, "magnetic_velocity_y": 0})
        magnetic.add_material("uniform_field", {"magnetic_conductivity": 0, "magnetic_current_density_external_real": 0,
                                                "magnetic_permeability": 1, "magnetic_remanence": 0,
                                                "magnetic_remanence_angle": 0, "magnetic_total_current_prescribed": 0,
                                                "magnetic_total_current_real": 0, "magnetic_velocity_angular": 0,
                                                "magnetic_velocity_x": 0, "magnetic_velocity_y": 0})
        magnetic.add_material("turn",
                              {"magnetic_conductivity": "58*1e6", "magnetic_current_density_external_real": "2*1e6",
                               "magnetic_permeability": 1, "magnetic_remanence": 0, "magnetic_remanence_angle": 0,
                               "magnetic_total_current_prescribed": 0, "magnetic_total_current_real": 0,
                               "magnetic_velocity_angular": 0, "magnetic_velocity_x": 0, "magnetic_velocity_y": 0})

        # geometry
        geometry = problem.geometry()
        geometry.add_edge(0, 0, 0.005, 0, boundaries={"magnetic": "Symmetry"})
        geometry.add_edge(0.005, 0, 0.005, 0.005)
        geometry.add_edge(0.005, 0.005, 0, 0.005)
        geometry.add_edge(0, 0.005, 0, 0, boundaries={"magnetic": "A=0"})
        geometry.add_edge("r1", 0, "r1+w", 0, boundaries={"magnetic": "Symmetry"})
        geometry.add_edge("r1+w", 0, "r1+w", "h")
        geometry.add_edge("r1+w", "h", "r1", "h")
        geometry.add_edge("r1", "h", "r1", 0)
        geometry.add_edge(0.005, 0, "r1", 0, boundaries={"magnetic": "Symmetry"})
        geometry.add_edge("r2", "2*h", "r2+w", "2*h")
        geometry.add_edge("r2+w", "2*h", "r2+w", "h+1e-5")
        geometry.add_edge("r2+w", "h+1e-5", "r2", "h+1e-5")
        geometry.add_edge("r2", "h+1e-5", "r2", "2*h")
        geometry.add_edge("r3", "2*h+1e-5", "r3+w", "2*h+1e-5")
        geometry.add_edge("r3+w", "2*h+1e-5", "r3+w", "3*h")
        geometry.add_edge("r3+w", "3*h", "r3", "3*h")
        geometry.add_edge("r3", "3*h", "r3", "2*h+1e-5")
        geometry.add_edge("r4", "3*h+2e-5", "r4+w", "3*h+2e-5")
        geometry.add_edge("r4+w", "3*h+2e-5", "r4+w", "4*h")
        geometry.add_edge("r4+w", "4*h", "r4", "4*h")
        geometry.add_edge("r4", "4*h", "r4", "3*h+2e-5")
        geometry.add_edge("r5", "4*h+2e-5", "r5+w", "4*h+2e-5")
        geometry.add_edge("r5", "4*h+2e-5", "r5", "5*h")
        geometry.add_edge("r5", "5*h", "r5+w", "5*h")
        geometry.add_edge("r5+w", "5*h", "r5+w", "4*h+2e-5")
        geometry.add_edge("r6", "5*h+2e-5", "r6+w", "5*h+2e-5")
        geometry.add_edge("r6+w", "5*h+2e-5", "r6+w", "6*h")
        geometry.add_edge("r6+w", "6*h", "r6", "6*h")
        geometry.add_edge("r6", "6*h", "r6", "5*h+2e-5")
        geometry.add_edge("r7", "6*h+2e-5", "r7+w", "6*h+2e-5")
        geometry.add_edge("r7+w", "6*h+2e-5", "r7+w", "7*h")
        geometry.add_edge("r7+w", "7*h", "r7", "7*h")
        geometry.add_edge("r7", "7*h", "r7", "6*h+2e-5")
        geometry.add_edge("r8", "7*h+2e-5", "r8+w", "7*h+2e-5")
        geometry.add_edge("r8+w", "7*h+2e-5", "r8+w", "8*h")
        geometry.add_edge("r8+w", "8*h", "r8", "8*h")
        geometry.add_edge("r8", "8*h", "r8", "7*h+2e-5")
        geometry.add_edge("r9", "8*h+2e-5", "r9+w", "8*h+2e-5")
        geometry.add_edge("r9+w", "8*h+2e-5", "r9+w", "9*h")
        geometry.add_edge("r9+w", "9*h", "r9", "9*h")
        geometry.add_edge("r9", "9*h", "r9", "8*h+2e-5")
        geometry.add_edge("r10", "9*h+2e-5", "r10+w", "9*h+2e-5")
        geometry.add_edge("r10+w", "9*h+2e-5", "r10+w", "10*h")
        geometry.add_edge("r10+w", "10*h", "r10", "10*h")
        geometry.add_edge("r10", "10*h", "r10", "9*h+2e-5")
        geometry.add_edge("r1+w", 0, "MW", 0, boundaries={"magnetic": "Symmetry"})
        geometry.add_edge("MW", 0, "MW", "MH", boundaries={"magnetic": "A=0"})
        geometry.add_edge("MW", "MH", 0, "MH", boundaries={"magnetic": "A=0"})
        geometry.add_edge(0, "MH", 0, 0.005, boundaries={"magnetic": "A=0"})

        geometry.add_label("0.8*MW", "0.8*MH", materials={"magnetic": "air"})
        geometry.add_label("r1+w/2", "0.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r2+w/2", "1.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r3+0.5*w", "2.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r4+0.5*w", "3.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r5+0.5*w", "4.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r6+w/2", "5.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r7+w/2", "6.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r8+w/2", "7.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r9+w/2", "8.5*h", materials={"magnetic": "turn"})
        geometry.add_label("r10+w/2", "9.5*h", materials={"magnetic": "turn"})
        geometry.add_label(0.0025, 0.0025, materials={"magnetic": "uniform_field"})

        # recipes
        magnetic.add_recipe_volume_integral("Wm", "magnetic_energy", [11], -1, -1)

        computation = problem.computation()
        computation.solve()

        solution = computation.solution("magnetic")

        B0 = 2e-3

        dxy = 0.5e-3
        nx = 8
        ny = 8
        dx = (5e-3 - dxy) / (nx - 1)
        dy = (5e-3 - dxy) / (ny - 1)

        f1 = 0.0
        f2 = 0.0
        f3 = 0.0
        for i in range(0, nx):
            xx = dxy + i * dx
            for j in range(0, ny):
                yy = dxy + j * dy

                point = solution.local_values(xx, yy)
                Br = point["Brr"]
                Bz = point["Brz"]

                Bp1s = math.sqrt((Br - 0.0) ** 2 + (Bz - B0) ** 2)
                f1 = max(f1, Bp1s)

                Bp2 = math.sqrt((Brp - Br) ** 2 + (Bzp - Bz) ** 2) + math.sqrt((Brm - Br) ** 2 + (Bzm - Bz) ** 2)
                f3 = max(f2, Bp2)

        f2 = sum(x) * 1e3

        return [f1, f2, f3]


def check_analytical_agros():
    vec = [0.00808, 0.0149, 0.00674, 0.0167, 0.00545, 0.0106, 0.0117, 0.0111, 0.01369, 0.00619]
    vec = [5.1e-3] * 10
    vec = [0.016932781137, 0.020220367784, 0.040397700883, 0.039838420509, 0.028051772403, 0.02469698998,
           0.024829362232, 0.014442787471, 0.024972441459, 0.037526654441]
    vec = [0.020798867779511312, 0.006138154021970008, 0.010012220217557183, 0.00690835649565163, 0.012212694015280686,
           0.00809716361914133, 0.018352792299856494, 0.00878331012406535, 0.02486595541682808, 0.00983825681954835]
    vec = [0.01020404406824542, 0.007528926158292741, 0.008708935266330736, 0.007283700503797654, 0.005215395421823215,
           0.01894915013137104, 0.020218991198109815, 0.005188444416722139, 0.015520989779088109, 0.0160617361729709]
    # vec = Individual([10e-3] * 10)
    ind = Individual(vec)

    problem_analytical = ProblemAnalytical()
    a = problem_analytical.evaluate(ind)
    problem_agros = AgrosSimple()
    n = problem_agros.evaluate(ind)

    print("anal ", a)
    print("agros", n)
    print("diff", math.fabs(a[0] - n[0]))


def optim_single():
    problem = ProblemAnalytical()  # AgrosSingle()
    # optimization
    algorithm = NSGAII(problem)
    algorithm.options['max_population_number'] = 100
    algorithm.options['max_population_size'] = 100
    algorithm.run()
    b = Results(problem)

    b.pareto_plot()

if __name__ == '__main__':
    # check_analytical_agros()
    optim_single()



import unittest
from artap.problem import Problem
from artap.individual import Individual
from artap.operators import CustomGenerator, LHSGenerator, Evaluator
from artap.algorithm import DummyAlgorithm
from artap.algorithm_sweep import SweepAlgorithm


class SweepProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "SweepProblem"
        self.parameters = [{'name': 'x_1', 'initial_value': 10, 'bounds': [-10, 30]}]
        self.costs = [{'name': 'F_1', 'criteria': 'minimize'}]

    def evaluate(self, individual: Individual):
        result = 0
        for i in individual.vector:
            result += i*i

        return [result]


class TestJob(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_sweep_evaluate_parallel(self):
        problem = SweepProblem()
        generator = LHSGenerator(problem.parameters)
        generator.init(4)

        algorithm = SweepAlgorithm(problem, generator=generator)
        algorithm.options['max_processes'] = 2
        algorithm.run()

        individuals = problem.populations[-1].individuals
        # values
        self.assertEqual(len(individuals), 4)
        self.assertTrue(problem.parameters[0]['bounds'][0] <= individuals[0].vector[0] <=
                        problem.parameters[0]['bounds'][1] and
                        problem.parameters[0]['bounds'][0] <= individuals[1].vector[0] <=
                        problem.parameters[0]['bounds'][1] and
                        problem.parameters[0]['bounds'][0] <= individuals[2].vector[0] <=
                        problem.parameters[0]['bounds'][1] and
                        problem.parameters[0]['bounds'][0] <= individuals[3].vector[0] <=
                        problem.parameters[0]['bounds'][1])

    def test_sweep_evaluate_serial(self):
        problem = SweepProblem()

        gen = CustomGenerator(problem.parameters)
        gen.init([[1, 2, 2], [3, 3, 2]])

        algorithm = SweepAlgorithm(problem, generator=gen)
        algorithm.run()

        individuals = problem.populations[-1].individuals
        self.assertEqual(individuals[0].costs, [9])
        self.assertEqual(individuals[1].costs, [22])

    def test_dummy_evaluate_serial(self):
        problem = SweepProblem()

        i1 = Individual([1, 2, 2])
        i2 = Individual([3, 3, 2])
        individuals = [i1, i2]

        algorithm = DummyAlgorithm(problem)
        algorithm.evaluate(individuals)

        self.assertEqual(individuals[0].costs, [9])
        self.assertEqual(individuals[1].costs, [22])

    def test_dummy_evaluate_parallel(self):
        problem = SweepProblem()

        i1 = Individual([1, 2, 2])
        i2 = Individual([3, 3, 2])
        individuals = [i1, i2]

        algorithm = DummyAlgorithm(problem)
        algorithm.options['max_processes'] = 2
        algorithm.evaluate(individuals)

        self.assertEqual(individuals[0].costs, [9])
        self.assertEqual(individuals[1].costs, [22])


if __name__ == '__main__':
    unittest.main()
